/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.curator.x.async.modeled;

/**
 * Serializing interface for models
 */
public interface ModelSerializer<T>
{
    /**
     * Given a model return the serialized bytes
     *
     * @param model model
     * @return bytes
     */
    byte[] serialize(T model);

    /**
     * Given bytes serialized via {@link #serialize(Object)} return
     * the model
     *
     * @param bytes serialized bytes
     * @return model
     * @throws RuntimeException if <code>bytes</code> is invalid or there was an error deserializing
     */
    T deserialize(byte[] bytes);

    /**
     * A pass through serializer
     */
    ModelSerializer<byte[]> raw = new ModelSerializer<byte[]>()
    {
        @Override
        public byte[] serialize(byte[] model)
        {
            return model;
        }

        @Override
        public byte[] deserialize(byte[] bytes)
        {
            return bytes;
        }
    };
}
